package Hexel.blocks;

import java.util.ArrayList;

import Hexel.blocks.rules.GrassGrowRule;
import Hexel.blocks.rules.HealRule;
import Hexel.blocks.rules.LightRule;
import Hexel.blocks.rules.SeedRule;
import Hexel.blocks.rules.TreeRules;
import Hexel.blocks.rules.WaterfallRule;
import Hexel.blocks.rules.WaterspreadRule;
import Hexel.blocks.types.Block;
import Hexel.blocks.types.BlockDirt;
import Hexel.blocks.types.BlockLeaf;
import Hexel.blocks.types.BlockSeed;
import Hexel.blocks.types.BlockTransparent;
import Hexel.blocks.types.BlockWater;
import Hexel.blocks.types.BlockWood;
import Hexel.chunk.Chunk;
import Hexel.chunk.Chunks;
import Hexel.math.HexGeometry;
import Hexel.math.Vector2i;
import Hexel.math.Vector3i;
import Hexel.util.Pair;

public class BlockRules {

	private Chunks chunks;

	ArrayList<BlockRule> everywhereRules = new ArrayList<BlockRule>();
	ArrayList<BlockRule> seedRules = new ArrayList<BlockRule>();
	ArrayList<BlockRule> woodRules = new ArrayList<BlockRule>();
	ArrayList<BlockRule> waterRules = new ArrayList<BlockRule>();
	ArrayList<BlockRule> dirtRules = new ArrayList<BlockRule>();
	ArrayList<BlockRule> leafRules = new ArrayList<BlockRule>();
	ArrayList<BlockRule> transparentRules = new ArrayList<BlockRule>();

	public BlockRules(Chunks chunks) {
		this.chunks = chunks;

		this.everywhereRules.add(new HealRule());
		this.transparentRules.add(new LightRule());
		
		this.seedRules.add(new SeedRule());
		TreeRules.addRules(this.woodRules, this.leafRules);
		this.waterRules.add(new WaterfallRule());
		this.waterRules.add(new WaterspreadRule());
		this.dirtRules.add(new GrassGrowRule());
		
		this.waterRules.addAll(this.transparentRules);
		this.leafRules.addAll(this.transparentRules);
		
		this.dirtRules.addAll(this.everywhereRules);
		this.seedRules.addAll(this.everywhereRules);
		this.woodRules.addAll(this.everywhereRules);
		this.waterRules.addAll(this.everywhereRules);
		this.leafRules.addAll(this.everywhereRules);
	}

	public Pair<Integer, ArrayList<BlockDelta>> apply(int x, int y, int z,
			Chunk c, Block b, int steps, Vector3i tmp, Vector2i tmp2) {
		ArrayList<BlockRule> rules = null;
		if (b instanceof BlockSeed) {
			rules = this.seedRules;
		} else if (b instanceof BlockWood) {
			rules = this.woodRules;
		} else if (b instanceof BlockWater) {
			rules = this.waterRules;
		} else if (b instanceof BlockLeaf) {
			rules = this.leafRules;
		} else if (b instanceof BlockDirt) {
			rules = this.dirtRules;
		} else if (b instanceof BlockTransparent) {
			rules = this.transparentRules;
		}
		ArrayList<BlockDelta> deltas;
		int minStepsUntilConsider = Integer.MAX_VALUE;
		if (rules == null)
			rules = this.everywhereRules;
		deltas = new ArrayList<BlockDelta>();
		boolean bail = false;
		int bx = x - c.cx*32;
		int by = y - c.cy*16;
		int bz = z - c.cz*32;
		boolean noMatches = true;
		for (BlockRule rule : rules) {
			int stepsUntilConsider = rule.stepsUntilConsider(x, y, z, b, c, this.chunks, steps, tmp, tmp2);
			if (stepsUntilConsider == 0 || c.stepsToSim[bx][by][bz] > 0) {
				boolean isMatch = rule.isMatch(x, y, z, b, c, this.chunks, steps, tmp, tmp2);
				if (isMatch) {
					noMatches = false;
					ArrayList<BlockDelta> additionalDeltas = rule.apply(x, y, z, b, c, this.chunks, steps, tmp, tmp2);
					if (additionalDeltas != null){
						for (BlockDelta delta : additionalDeltas) {
							if (delta.immediate) {
								this.chunks.setBlock(delta.x, delta.y, delta.z, delta.block, tmp, tmp2);
								HexGeometry.globalToLocal(delta.x, delta.y, delta.z, tmp);
								c.stepsToSim[tmp.x][tmp.y][tmp.z] = delta.stepsToSim-1;
							}
						}
						deltas.addAll(additionalDeltas);
						for (BlockDelta delta : additionalDeltas) {
							if (delta.immediate && delta.x == x && delta.y == y && delta.z == z){
								bail = true;
								break;
							}
						}
					}
					int nextStepsUntilConsider = rule.stepsUntilConsider(x, y, z,
							b, c, this.chunks, steps + 1, tmp, tmp2);
					minStepsUntilConsider = Math.min(minStepsUntilConsider, nextStepsUntilConsider);
					if (bail)
						break;
				}
			} else {
				stepsUntilConsider = rule.stepsUntilConsider(x, y, z, b, c, this.chunks, steps + 1, tmp, tmp2);
				minStepsUntilConsider = Math.min(minStepsUntilConsider, stepsUntilConsider);
			}
		}
		if (rules.size() == 0){
			if (c.stepsToSim[bx][by][bz] > 0)
				c.stepsToSim[bx][by][bz] = 0;
		}
		if (noMatches){
			c.stepsToSim[bx][by][bz] = 0;
		}
		if (c.stepsToSim[bx][by][bz] > 0){
			if (!bail)
				c.stepsToSim[bx][by][bz] -= 1;
			minStepsUntilConsider = -1;
		}
		return new Pair<Integer, ArrayList<BlockDelta>>(minStepsUntilConsider, deltas);
	}
}
