package Hexel.chunk;

import java.util.concurrent.ExecutionException;

import Hexel.Cleanup;
import Hexel.blocks.types.Block;
import Hexel.blocks.types.BlockTransparent;
import Hexel.generation.ChunkGenerator;
import Hexel.math.Vector2i;
import Hexel.math.Vector3i;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.cache.RemovalListener;
import com.google.common.cache.RemovalNotification;

public class Chunks {

	private ChunkGenerator chunkGenerator;
	private LoadingCache<Vector3i, Chunk> chunkCache;
	private LoadingCache<Vector2i, Integer[][]> highestBlockCache;

	public Chunks(Cleanup cleanup, int numToBuffer) {
		this.chunkCache = CacheBuilder.newBuilder().maximumSize(numToBuffer)
				.recordStats()
				.removalListener(new RemovalListener<Vector3i, Chunk>() {
					@Override
					public void onRemoval(
							RemovalNotification<Vector3i, Chunk> notification) {
						Chunk c = notification.getValue();
						if (c != null && c.modified) {
							ChunkFile.save(c);
						}
					}
				}).build(new CacheLoader<Vector3i, Chunk>() {
					@Override
					public Chunk load(Vector3i p) {
						Chunk chunk;
						if (ChunkFile.has(p)) {
							chunk = ChunkFile.load(p);
						} else {
							chunk = Chunks.this.chunkGenerator.genChunk(p);
						}
						return chunk;
					}
				});
		this.highestBlockCache = CacheBuilder.newBuilder().maximumSize(numToBuffer)
				.removalListener(new RemovalListener<Vector2i, Integer[][]>() {
					@Override
					public void onRemoval(
							RemovalNotification<Vector2i, Integer[][]> notification) {
						Integer[][] map = notification.getValue();
						Vector2i loc = notification.getKey();
						if (map != null) {
							HighestBlockFile.save(loc, map);
						}
					}
				}).build(new CacheLoader<Vector2i, Integer[][]>() {
					@Override
					public Integer[][] load(Vector2i p) {
						Integer[][] map;
						if (HighestBlockFile.has(p)) {
							map = HighestBlockFile.load(p);
						} else {
							map = new Integer[Chunk.WIDTH][Chunk.HEIGHT];
							for (int x = 0; x < Chunk.WIDTH; x++){
								for (int y = 0; y < Chunk.HEIGHT; y++){
									map[x][y] = Integer.MIN_VALUE;
								}
							}
						}
						return map;
					}
				});

		this.chunkGenerator = new ChunkGenerator(this, cleanup);

		cleanup.add(new Runnable() {
			@Override
			public void run() {
				unloadAllChunks();
			}
		});
	}

	public Chunk getChunk(Vector3i cpos) {
		try {
			Chunk c = this.chunkCache.getIfPresent(cpos);
			if (c != null)
				return c;
			else 
				return this.chunkCache.get(new Vector3i(cpos));
		} catch (Exception e) {
			System.out.println("Caught Exception retrieving chunk from cache!");
			e.printStackTrace();
			System.exit(1);
			return null;
		}
	}

	public void unloadAllChunks() {
		this.chunkCache.invalidateAll();
		this.highestBlockCache.invalidateAll();
	}

	public int getHighestBlockAtXY(int x, int y, Vector2i pos){
		int cx = (int) Math.floor(x / 32.0);
		int cy = (int) Math.floor(y / 16.0);

		int bx = x - cx * 32;
		int by = y - cy * 16;
		pos.x = cx;
		pos.y = cy;
		Integer[][] heights;
		try {
			heights = this.highestBlockCache.get(pos);
			return heights[bx][by];
		} catch (ExecutionException e) {
			e.printStackTrace();
			System.exit(1);
		}
		return Integer.MIN_VALUE;
	}

	private void updateHighestBlockAtXY(int x, int y, int z, Vector2i pos){
		int cx = (int) Math.floor(x / 32.0);
		int cy = (int) Math.floor(y / 16.0);

		int bx = x - cx * 32;
		int by = y - cy * 16;
		pos.x = cx;
		pos.y = cy;
		Integer[][] heights;
		try {
			heights = this.highestBlockCache.get(pos);
			if (z > heights[bx][by])
				heights[bx][by] = z;
		} catch (ExecutionException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	public Block getBlock(int x, int y, int z, Vector3i p) {
		int cx = (int) Math.floor(x / 32.0);
		int cy = (int) Math.floor(y / 16.0);
		int cz = (int) Math.floor(z / 32.0);

		int bx = x - cx * 32;
		int by = y - cy * 16;
		int bz = z - cz * 32;

		p.x = cx;
		p.y = cy;
		p.z = cz;
		Chunk chunk = getChunk(p);
		return chunk.get(bx, by, bz);
	}

	public Block getBlock(int cx, int cy, int cz, int x, int y, int z, Vector3i p) {
		int gx = cx * 32 + x;
		int gy = cy * 16 + y;
		int gz = cz * 32 + z;

		return this.getBlock(gx, gy, gz, p);
	}

	public void awakenBlock(int x, int y, int z, Vector3i tmp) {
		int cx = (int) Math.floor(x / 32.0);
		int cy = (int) Math.floor(y / 16.0);
		int cz = (int) Math.floor(z / 32.0);

		tmp.x = cx;
		tmp.y = cy;
		tmp.z = cz;
		Chunk chunk = getChunk(tmp);

		chunk.stepsToNeedSim = 0;
	}
	
	public void setStepsToSim(int x, int y, int z, int stepsToSim, Vector3i p){
		int cx = (int) Math.floor(x / 32.0);
		int cy = (int) Math.floor(y / 16.0);
		int cz = (int) Math.floor(z / 32.0);

		int bx = x - cx * 32;
		int by = y - cy * 16;
		int bz = z - cz * 32;

		p.x = cx;
		p.y = cy;
		p.z = cz;
		Chunk chunk = getChunk(p);
		chunk.stepsToSim[bx][by][bz] = stepsToSim;
		if (stepsToSim > 0)
			chunk.stepsToNeedSim = -1;
	}

	public void setBlock(int x, int y, int z, Block b, Vector3i p, Vector2i tmp2) {
		int cx = (int) Math.floor(x / 32.0);
		int cy = (int) Math.floor(y / 16.0);
		int cz = (int) Math.floor(z / 32.0);

		int bx = x - cx * 32;
		int by = y - cy * 16;
		int bz = z - cz * 32;

		p.x = cx;
		p.y = cy;
		p.z = cz;


		Chunk chunk = getChunk(p);

		chunk.set(bx, by, bz, b);
		chunk.dirty = true;
		chunk.stepsToNeedSim = 0;

		if (! (b instanceof BlockTransparent)){
			updateHighestBlockAtXY(x, y, z, tmp2);
		}

		chunk.modified = true;

		for (int dx = -1; dx <= 1; dx++) {
			for (int dy = -1; dy <= 1; dy++) {
				for (int dz = -1; dz <= 1; dz++) {
					int x2 = x + dx;
					int y2 = y + dy;
					int z2 = z + dz;

					p.x = (int) Math.floor(x2 / 32.0);
					p.y = (int) Math.floor(y2 / 16.0);
					p.z = (int) Math.floor(z2 / 32.0);

					if (p.x != cx || p.y != cy || p.z != cz) {
						chunk = getChunk(p);
						chunk.dirty = true;
						chunk.stepsToNeedSim = 0;
					}
				}
			}
		}
	}
}
