package Hexel.rendering;

import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;

import javax.imageio.ImageIO;
import javax.media.opengl.GL;
import javax.media.opengl.GL2;
import javax.media.opengl.GL2ES1;
import javax.media.opengl.fixedfunc.GLPointerFunc;

import Hexel.chunk.Chunk;
import Hexel.chunk.Chunks;
import Hexel.math.HexGeometry;
import Hexel.math.Vector2d;
import Hexel.math.Vector3i;

import com.jogamp.opengl.util.texture.Texture;
import com.jogamp.opengl.util.texture.TextureData;
import com.jogamp.opengl.util.texture.TextureIO;
import com.jogamp.opengl.util.texture.awt.AWTTextureIO;

public class ChunkRenderer {

	private Chunks chunks;

	private Texture ttex;

	private LinkedBlockingQueue<GLChunk> glChunksToLoad = new LinkedBlockingQueue<GLChunk>();
	private HashMap<Vector3i, GLChunk> glChunkTable = new HashMap<Vector3i, GLChunk>();

	public HashMap<Vector3i, GLChunk> getGLChunkTable() {
		return this.glChunkTable;
	}

	private ExecutorService chunkVertexBufferGeneratorThreadPool = Executors
			.newFixedThreadPool(8);

	public ChunkRenderer(Chunks chunks) {
		this.chunks = chunks;
	}

	public void init(GL2 gl) {
		try {
			this.loadTex(gl);
		} catch(IOException e){
			System.out.println("unable to load atlas");
			e.printStackTrace();
			System.exit(1);
		}

		gl.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER,
				GL.GL_NEAREST);
		gl.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER,
				GL.GL_NEAREST);
	}
	
	private void loadTex(GL2 gl) throws IOException{
		BufferedImage atlas = TextureAtlas.createAtlas();
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ImageIO.write(atlas, "png", baos);
		InputStream is = new ByteArrayInputStream(baos.toByteArray());
		TextureData data = TextureIO.newTextureData(gl.getGLProfile(), is, false, "png");
		this.ttex = TextureIO.newTexture(data);
	}

	public void loadChunks(final ArrayList<Vector3i> points) {
		if (points.size() == 0)
			return;
		for (Vector3i p : points) {
			if (!hasChunk(p))
				this.glChunkTable.put(p, null);
		}
		this.chunkVertexBufferGeneratorThreadPool.execute(new Runnable() {
			@Override
			public void run() {
				for (Vector3i point : points) {
					Chunk chunk = ChunkRenderer.this.chunks.getChunk(point);
					GLChunk glChunk = new GLChunk(chunk, ChunkRenderer.this.chunks);
					glChunk.position = new Vector3i(chunk.cx, chunk.cy,
							chunk.cz);
					glChunk.genArrayList();
					ChunkRenderer.this.loadGLChunk(glChunk);
				}
			}
		});
	}

	public boolean hasChunk(Vector3i pos) {
		return this.glChunkTable.containsKey(pos);
	}

	public void loadGLChunk(GLChunk glChunk) {
		this.glChunksToLoad.add(glChunk);
	}

	public void unloadGLChunk(Vector3i pos) {
		this.glChunkTable.remove(pos);
	}

	public void update(GL2 gl) {
		int i = 0;
		while (!this.glChunksToLoad.isEmpty()) {
			GLChunk glChunk = this.glChunksToLoad.poll();
			glChunk.initBuffer(gl);
			this.glChunkTable.put(glChunk.position, glChunk);
			i++;
			if (i > 10)
				break;
		}
	}

	public void render(GL2 gl) {

		this.ttex.enable(gl);
		this.ttex.bind(gl);

		Iterator<Map.Entry<Vector3i, GLChunk>> it = this.glChunkTable.entrySet().iterator();

		while (it.hasNext()) {
			Map.Entry<Vector3i, GLChunk> entry = it.next();
			Vector3i position = entry.getKey();
			GLChunk glChunk = entry.getValue();

			if (glChunk == null || glChunk.buffer == null)
				continue;

			gl.glPushMatrix();

			Vector2d hexPosition = new Vector2d();
			HexGeometry.hexToCartesian(position.x, position.y, hexPosition);

			gl.glTranslatef((float) (hexPosition.x * 16),
					(float) (hexPosition.y * 16), position.z * 16);

			gl.glBindBuffer(GL.GL_ARRAY_BUFFER, glChunk.buffer.id[0]);

			// SOLIDS
			gl.glAlphaFunc(GL.GL_EQUAL, 1.0f);
			gl.glEnable(GL2ES1.GL_ALPHA_TEST);

			gl.glEnableClientState(GLPointerFunc.GL_VERTEX_ARRAY);
			gl.glEnableClientState(GLPointerFunc.GL_TEXTURE_COORD_ARRAY);

			gl.glVertexPointer(3, GL.GL_FLOAT, 9 * 4, 0);
			gl.glTexCoordPointer(2, GL.GL_FLOAT, 9 * 4, 3 * 4);

			gl.glDrawArrays(GL.GL_TRIANGLES, 0, glChunk.buffer.size / (4 * 9));

			gl.glDisableClientState(GLPointerFunc.GL_VERTEX_ARRAY);
			gl.glDisableClientState(GLPointerFunc.GL_TEXTURE_COORD_ARRAY);

			gl.glDisable(GL2ES1.GL_ALPHA_TEST);

			gl.glPopMatrix();

		}
		gl.glEnable(GL.GL_BLEND);
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);

		it = this.glChunkTable.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry<Vector3i, GLChunk> entry = it.next();
			Vector3i position = entry.getKey();
			GLChunk glChunk = entry.getValue();

			if (glChunk == null || glChunk.buffer == null)
				continue;

			gl.glPushMatrix();

			Vector2d hexPosition = new Vector2d();
			HexGeometry.hexToCartesian(position.x, position.y, hexPosition);

			gl.glTranslatef((float) (hexPosition.x * 16),
					(float) (hexPosition.y * 16), position.z * 16);

			gl.glBindBuffer(GL.GL_ARRAY_BUFFER, glChunk.buffer.id[0]);

			gl.glEnableClientState(GLPointerFunc.GL_VERTEX_ARRAY);
			gl.glEnableClientState(GLPointerFunc.GL_COLOR_ARRAY);

			gl.glVertexPointer(3, GL.GL_FLOAT, 9 * 4, 0);
			gl.glColorPointer(4, GL.GL_FLOAT, 9 * 4, 5 * 4);

			gl.glDrawArrays(GL.GL_TRIANGLES, 0, glChunk.buffer.size / (4 * 9));

			gl.glDisableClientState(GLPointerFunc.GL_COLOR_ARRAY);
			gl.glDisableClientState(GLPointerFunc.GL_VERTEX_ARRAY);

			gl.glPopMatrix();

		}

		it = this.glChunkTable.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry<Vector3i, GLChunk> entry = it.next();
			Vector3i position = entry.getKey();
			GLChunk glChunk = entry.getValue();

			if (glChunk == null || glChunk.buffer == null)
				continue;

			gl.glPushMatrix();

			Vector2d hexPosition = new Vector2d();
			HexGeometry.hexToCartesian(position.x, position.y, hexPosition);

			gl.glTranslatef((float) (hexPosition.x * 16),
					(float) (hexPosition.y * 16), position.z * 16);

			gl.glBindBuffer(GL.GL_ARRAY_BUFFER, glChunk.buffer.id[0]);

			// NOT SOLIDS
			gl.glAlphaFunc(GL.GL_NOTEQUAL, 1.0f);
			gl.glEnable(GL2ES1.GL_ALPHA_TEST);

			gl.glEnableClientState(GLPointerFunc.GL_VERTEX_ARRAY);
			gl.glEnableClientState(GLPointerFunc.GL_TEXTURE_COORD_ARRAY);

			gl.glVertexPointer(3, GL.GL_FLOAT, 9 * 4, 0);
			gl.glTexCoordPointer(2, GL.GL_FLOAT, 9 * 4, 3 * 4);

			gl.glDrawArrays(GL.GL_TRIANGLES, 0, glChunk.buffer.size / (4 * 9));

			gl.glDisableClientState(GLPointerFunc.GL_VERTEX_ARRAY);
			gl.glDisableClientState(GLPointerFunc.GL_TEXTURE_COORD_ARRAY);

			gl.glDisable(GL2ES1.GL_ALPHA_TEST);
			gl.glPopMatrix();

		}
		gl.glDisable(GL.GL_BLEND);
		this.ttex.disable(gl);
	}
}
